package java.util.concurrent;

import java.security.ProtectionDomain;
import sun.misc.Unsafe;
import java.security.AccessControlContext;

public class ForkJoinWorkerThread extends Thread
{
    final ForkJoinPool pool;
    final ForkJoinPool.WorkQueue workQueue;
    private static final AccessControlContext INNOCUOUS_ACC;
    private static final Unsafe U;
    private static final long THREADLOCALS;
    private static final long INHERITABLETHREADLOCALS;
    private static final long INHERITEDACCESSCONTROLCONTEXT;
    
    protected ForkJoinWorkerThread(ForkJoinPool pool) {
        super("aForkJoinWorkerThread");
        this.pool = pool;
        this.workQueue = pool.registerWorker(this);
    }
    
    ForkJoinWorkerThread(ForkJoinPool pool, Object o) {
        super("aForkJoinWorkerThread");
        ForkJoinWorkerThread.U.putOrderedObject(this, ForkJoinWorkerThread.INHERITEDACCESSCONTROLCONTEXT, ForkJoinWorkerThread.INNOCUOUS_ACC);
        this.pool = pool;
        this.workQueue = pool.registerWorker(this);
    }
    
    ForkJoinWorkerThread(ForkJoinPool pool, ThreadGroup group, AccessControlContext x) {
        super(group, null, "aForkJoinWorkerThread");
        ForkJoinWorkerThread.U.putOrderedObject(this, ForkJoinWorkerThread.INHERITEDACCESSCONTROLCONTEXT, x);
        this.eraseThreadLocals();
        this.pool = pool;
        this.workQueue = pool.registerWorker(this);
    }
    
    public ForkJoinPool getPool() {
        return this.pool;
    }
    
    public int getPoolIndex() {
        return this.workQueue.getPoolIndex();
    }
    
    protected void onStart() {
    }
    
    protected void onTermination(Throwable t) {
    }
    
    @Override
    public void run() {
        if (this.workQueue.array == null) {
            Throwable t = null;
            try {
                this.onStart();
                this.pool.runWorker(this.workQueue);
                try {
                    this.onTermination(t);
                }
                catch (Throwable t2) {
                    if (t == null) {
                        t = t2;
                    }
                }
                finally {
                    this.pool.deregisterWorker(this, t);
                }
            }
            catch (Throwable t3) {
                t = t3;
                try {
                    this.onTermination(t);
                }
                catch (Throwable t4) {
                    if (t == null) {
                        t = t4;
                    }
                }
                finally {
                    this.pool.deregisterWorker(this, t);
                }
            }
            finally {
                try {
                    this.onTermination(t);
                }
                catch (Throwable t5) {
                    if (t == null) {
                        t = t5;
                    }
                }
                finally {
                    this.pool.deregisterWorker(this, t);
                }
            }
        }
    }
    
    final void eraseThreadLocals() {
        ForkJoinWorkerThread.U.putObject(this, ForkJoinWorkerThread.THREADLOCALS, null);
        ForkJoinWorkerThread.U.putObject(this, ForkJoinWorkerThread.INHERITABLETHREADLOCALS, null);
    }
    
    void afterTopLevelExec() {
    }
    
    static {
        INNOCUOUS_ACC = new AccessControlContext(new ProtectionDomain[] { new ProtectionDomain(null, null) });
        try {
            U = Unsafe.getUnsafe();
            Class<Object> clazz = Thread.class;
            THREADLOCALS = ForkJoinWorkerThread.U.objectFieldOffset(clazz.getDeclaredField("threadLocals"));
            INHERITABLETHREADLOCALS = ForkJoinWorkerThread.U.objectFieldOffset(clazz.getDeclaredField("inheritableThreadLocals"));
            INHERITEDACCESSCONTROLCONTEXT = ForkJoinWorkerThread.U.objectFieldOffset(clazz.getDeclaredField("inheritedAccessControlContext"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class InnocuousForkJoinWorkerThread extends ForkJoinWorkerThread
    {
        private static final ThreadGroup innocuousThreadGroup;
        
        InnocuousForkJoinWorkerThread(ForkJoinPool forkJoinPool) {
            super(forkJoinPool, ForkJoinWorkerThread.InnocuousForkJoinWorkerThread.innocuousThreadGroup, ForkJoinWorkerThread.INNOCUOUS_ACC);
        }
        
        @Override
        void afterTopLevelExec() {
            this.eraseThreadLocals();
        }
        
        @Override
        public ClassLoader getContextClassLoader() {
            return ClassLoader.getSystemClassLoader();
        }
        
        @Override
        public void setUncaughtExceptionHandler(Thread.UncaughtExceptionHandler uncaughtExceptionHandler) {
        }
        
        @Override
        public void setContextClassLoader(ClassLoader classLoader) {
            throw new SecurityException("setContextClassLoader");
        }
        
        private static ThreadGroup createThreadGroup() {
            try {
                Unsafe unsafe = Unsafe.getUnsafe();
                Class<Object> clazz = Thread.class;
                Class<Object> clazz2 = ThreadGroup.class;
                long objectFieldOffset = unsafe.objectFieldOffset(clazz.getDeclaredField("group"));
                long objectFieldOffset2 = unsafe.objectFieldOffset(clazz2.getDeclaredField("parent"));
                ThreadGroup threadGroup = (ThreadGroup)unsafe.getObject(Thread.currentThread(), objectFieldOffset);
                while (threadGroup != null) {
                    ThreadGroup threadGroup2 = (ThreadGroup)unsafe.getObject(threadGroup, objectFieldOffset2);
                    if (threadGroup2 == null) {
                        return new ThreadGroup(threadGroup, "InnocuousForkJoinWorkerThreadGroup");
                    }
                    threadGroup = threadGroup2;
                }
            }
            catch (Exception cause) {
                throw new Error(cause);
            }
            throw new Error("Cannot create ThreadGroup");
        }
        
        static {
            innocuousThreadGroup = createThreadGroup();
        }
    }
}
